<?php

namespace Drupal\mcapi_credcom;

use Drupal\mcapi\Entity\Currency;
use CreditCommons\Workflow;
use CreditCommons\Transaction as CCTransaction;
use Drupal\mcapi_credcom\Functions;

/**
 * An object to represent the credit commons data which doesn't fit in the transaction.
 */
class CredCom {

  const MCAPI_CREDCOM_STATE_MAP = [
    #drupal => credcom
    'validated' => Workflow::STATE_VALIDATED,
    'pending' => Workflow::STATE_PENDING,
    'done' => Workflow::STATE_COMPLETED,
    'erased' => Workflow::STATE_ERASED,
  ];

  /**
   * 1 capitalised PVC or E
   * @var string
   */
  public string $state;

  /**
   * @var int
   */
  public int $version;

  /**
   * Additional entries.
   * @var array
   */
  public array $additional = [];

  /**
   * Constructor
   * NB we don't store database as a property because such objects are hard to serialize.
   */
  public function __construct(
    public string $uuid,
    public Workflow $workflow,
    public array $entries,
    public string $curr_id,
    public ?int $xid = 0,
    ) {
    $this->state = 'init';
    $this->version = -1;
  }

  static function load($uuid) {
    $result = \Drupal::database()->select('mcapi_credcom_data', 'e')
      ->fields('e', ['xid', 'uuid', 'version', 'state', 'workflow', 'entries', 'curr_id'])
      ->condition('uuid', $uuid, 'IN')
      ->execute()->fetch();
    if ($result) {
      $credcom = new static(
        $result->uuid,
        new Workflow($result->workflow),
        unserialize($result->entries),
        $result->curr_id
      );
      $credcom->state = $result->state;
      $credcom->version = $result->version;
      return $credcom;
    }
  }

  /**
   * Called in hook_mcapi_transaction_presave().
   * @param string $credcom_state
   * @param ?bool $relay
   * @return bool
   *    TRUE on success
   */
  function changeState(string $credcom_state, $relay = FALSE) : bool {
    $this->version++;
    $this->state = $credcom_state;
    if ($relay) {
      $currency = Currency::load($this->curr_id);
      $requester = new TrunkwardNodeRequester($currency);
      try {
        $requester->transactionChangeState($this->uuid, $credcom_state);
      }
      catch(\CreditCommons\Exceptions\CCError $e) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Called via hook_mcapi_transaction_update().
   */
  function update() {
    \Drupal::database()->update('mcapi_credcom_data')
      ->fields(['state'=> $this->state, 'version' => $this->version])
      ->condition('uuid', $this->uuid)
      ->execute();
    $this->saveHash();
  }

  /**
   * @param int $xid
   */
  function insert($xid) {
    \Drupal::database()->insert('mcapi_credcom_data')
      ->fields([
        'xid' => $xid,
        'uuid' => $this->uuid,
        'state'=> $this->state,
        'workflow'=> strval($this->workflow),
        'version' => $this->version,
        'entries' => serialize($this->entries),
        'curr_id' => $this->curr_id
      ])
      ->execute();
    $this->saveHash();
  }

  /**
   *
   */
  function saveHash() {
    if ($this->version) {
      $hash_string = $this->getHashString();
      \Drupal::database()->insert('mcapi_credcom_hashchain')
        ->fields([
          'uuid' => $this->uuid,
          'curr_id' => $this->curr_id,
          'version' => $this->version,
          'hash' => md5($hash_string),
          'hash_string' => $hash_string
        ])
        ->execute();
    }
  }

  /**
   * @todo replace this with the cc-php-lib
   */
  private function getHashString() : string {
    $rows = [];
    $node_name = Currency::load($this->curr_id)->getThirdPartySetting('mcapi_credcom', 'node_name');
    foreach ($this->entries as $e) {
      if (Functions::credcomPathIsLocal($e->payee, $node_name) xor Functions::credcomPathIsLocal($e->payer, $node_name)) {
        // All quants are trunkwardQuants
        $rows[] = $e->description .'@'.$e->quant;
      }
    }
    return join('|', [
      $this->uuid,
      strval($this->workflow),
      join(';', $rows),
      $this->state,
      $this->version,
      Functions::getlastHash($this->curr_id),
    ]);
  }

  function getTransaction() : CCTransaction {
    return CCTransaction::create((object)(array)$this);
  }

}
