<?php

namespace Drupal\mcapi_credcom;

use Drupal\mcapi\Mcapi;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi_credcom\CredCom;
use Drupal\mcapi_credcom\EventSubscriber\ExceptionSubscriber;
use Drupal\mcapi_credcom\Functions;
use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Workflow;
use CreditCommons\Exceptions\WorkflowViolation;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\TempStore\SharedTempStore;

/**
 * Handle all incoming requests from the trunkward node.
 */
class EndPoint extends \Drupal\Core\Controller\ControllerBase {

  private $logger;
  private int $intertradingWid;
  private $tempstore;
  private Currency $currency;

  /**
   * See HashChecker for more set up operations.
   */
  public function __construct(Request $request, SharedTempStore $credcom_tempstore) {
    $this->currency = $request->attributes->get('mcapi_currency');
    global $error_context;
    Functions::createErrorContext($this->currency);

    $this->intertradingWid = $this->config('mcapi_credcom.settings')->get('wallet_id');
    \Drupal::getContainer()->set('exception.final', new ExceptionSubscriber($this->configFactory));
    $this->logger = $this->getLogger('mcapi_credcom');
    $this->logger->debug('@method @path<br />@body', ['@method' => $request->getmethod(), '@path' => $request->getUri(), '@body' => strval($request->getContent())]);
    $this->tempstore = $credcom_tempstore;
  }

  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('tempstore.shared')->get('mcapi_credcom')
    );
  }

  /**
   * Return the exchange rate.
   */
  function getAbout() {
    [$balances_asc, $trades, $volumes, $incomes, $spending] = Mcapi::tradingStats($this->currency, $conditions = []);

    $curr_settings = $this->currency->getThirdPartySettings('mcapi_credcom');
    $all_wallet_ids = \Drupal::entityQuery('mcapi_wallet')->accessCheck()->execute();

    $about = [
      'format' => $curr_settings['local_curr_symbol'],
      'rate' => $curr_settings['rate'],
      'absolute_path' => $this->currency->getThirdpartySetting('mcapi_credcom', 'absolute_path') .'/'.$this->currency->getThirdpartySetting('mcapi_credcom', 'node_name'),
      'validated_window' => '100000', // Not implemented
      'accounts' => count($all_wallet_ids),
      'traders' => count($balances_asc),
      'trades' => array_sum($trades)/2,
      'volume' => array_sum($volumes)/2,
    ];
    return JsonResponse::fromJsonString(json_encode($about));
  }

  /**
   * @param Request $request
   * @return JsonResponse
   */
  function autocompleteNames(Request $request) {
    $fragment = $request->query->get('acc_path');
    $query = \Drupal::entityQuery('mcapi_wallet')
      ->accessCheck(TRUE)
      ->range(0, 10);
    $remote_settings = $this->currency->getThirdPartySettings('mcapi_credcom');
    $fragment = str_replace($remote_settings['absolute_path'].'/', '', $fragment);
    [$node_name, $q] = explode('/', $fragment.'/');
    if ($q) {
      $query->condition('name', '%'.$q.'%', 'LIKE');
    }
    $wids = $query->execute();
    $result = [];
    $abs_prefix = $remote_settings['absolute_path'] .'/'. $remote_settings['node_name'];
    foreach (Wallet::loadMultiple($wids) as $wid => $wallet) {
      // Credit commons wallet names can't have spaces.
      $result[] = $abs_prefix .'/'. str_replace(' ', '-', $wallet->name->value);
    }
    return JsonResponse::fromJsonString(json_encode(['data' => $result]));
  }

  /**
   * Handle incoming transaction request.
   * @return JsonResponse
   */
  function relayTransaction(Request $request) {
    $contents = $request->getContent();
    $this->logger->debug('Received relayed json @credcom', ['@credcom' => $contents]);
    $credComFull = json_decode($contents);
    foreach ($credComFull->entries as &$entry) {
      $entry = Functions::upcastEntry($entry, $this->currency->getThirdPartySetting('mcapi_credcom', 'node_name'));
    }
    $credCom = new CredCom(
      $credComFull->uuid,
      new Workflow($credComFull->workflow),
      $credComFull->entries,
      $this->currency->id()
    );
    $transaction = Functions::convertRelayedTransactionToDrupal($credCom, $this->currency); // includes validation
    $transaction->credCom = $credCom;
    if ($credCom->workflow->requiresConfirmation()) {
      $credCom->changeState(Workflow::STATE_VALIDATED);
      $this->tempstore->set($credCom->uuid . $transaction->credCom->curr_id, $transaction);
      $response_code = 200;
    }
    else {
      $start_state = $credCom->workflow->startState();
      $credCom->changeState($start_state);
      $transaction->state->setValue(array_search($start_state, CredCom::MCAPI_CREDCOM_STATE_MAP));
      $credCom->version = 1; //ready to be saved to version 1
      $transaction->save();
      $response_code = 201;
    }
    $payload = ['data' => $transaction->credCom->additional, 'meta' => ['secs_valid_left' => 300]];
    return JsonResponse::fromJsonString(json_encode($payload), $response_code);
  }

  /**
   *
   * @param Currency $currency
   * @param string $uuid
   * @param string $cc_state
   * @return JsonResponse
   * @throws WorkflowViolation
   */
  function stateChange(string $uuid, string $cc_state) {
    $node_name = $this->currency->getThirdPartySetting('mcapi_credcom', 'node_name');
    if ($transaction = $this->tempstore->get($uuid . $node_name)) {
      $this->tempstore->delete($uuid . $node_name);
    }
    else {
      $query = \Drupal::database()->select('mcapi_credcom_data', 'cc')
        ->fields('t', ['serial'])
        ->condition('t.uuid', $uuid)
        ->condition('cc.currency_id', $this->currency->id());
      $query->join('mcapi_transaction', 't', 't.xid = cc.xid');
      $serial = $query->execute()->fetchField();
      $transaction = \Drupal\mcapi\Entity\Transaction::load($serial);
      $this->logger->debug('Loaded @t', ['@t' => $transaction->toArray()]);
    }
    if (!isset($transaction->credCom)) {
      throw new CCFailure("Transaction $uuid seems not to be a remote transaction.");
    }
    $credCom = $transaction->credCom;

    $permitted = $credCom->workflow->canTransitionToState(
      $credCom->getTransaction(),
      $credCom->curr_id,
      $cc_state,
      \Drupal::currentUser()->hasPermission('manage mcapi')
    );

    if (!$permitted) {
      throw new WorkflowViolation(type: $credCom->workflow, from: $credCom->state, to: $cc_state);
    }
    // Run the remote operation first.
    $credCom->changeState($cc_state);
    $transaction->state->setValue(array_search($cc_state, CredCom::MCAPI_CREDCOM_STATE_MAP));
    if ($cc_state == Workflow::STATE_PENDING) {// Equivalent to mcapi_signatures_mcapi_entity_presave()
      $local_wallet = ($transaction->payee->target_id == $this->intertradingWid) ?
        $transaction->payer->entity :
        $transaction->payee->entity;
      $transaction->signatures[$local_wallet->getOwner()->id()] = 0;
    }
    elseif ($cc_state == Workflow::STATE_COMPLETED and $transaction->signatures) { // Equivalent to saying sign all signatures.
      mcapi_signature_service($transaction)->signOff();
    }

    $transaction->save();
    $this->logger->debug('Saved @uuid (@currency) to @cc_state', ['@uuid' => $uuid, '@currency' => $this->currency->id(), '@cc_state' => $cc_state]);
    return JsonResponse::fromJsonString('{"data": []}', 201);
  }


  /**
   *
   */
  private function accountSummary() {
    return JsonResponse::fromJsonString('["todo"]');
  }


}
