<?php

namespace Drupal\mcapi_credcom;

use CreditCommons\Entry;
use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Exceptions\CCError;
use CreditCommons\Workflow;
use Drupal\mcapi_credcom\CredCom;
use Drupal\mcapi_credcom\Account;
use Drupal\mcapi_credcom\AccountRemote;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\Storage\WalletStorage;
use Drupal\Core\Render\Markup;

/**
 * Access check for requests from trunkwards nodes.
 */
class Functions {

  static $remoteAccPath;

  /**
   * @param Transaction $transaction
   * @return bool
   *   TRUE if either the payer or payee is the balance of trade wallet.
   */
  static function isRemote(Transaction $transaction) : bool {
    $bot_wid = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
    return $transaction->payee->target_id == $bot_wid or $transaction->payer->target_id == $bot_wid;
  }

  /**
   * Convert the given credcom transaction into a Drupal transaction.
   * The given entries are converted to children and any children added by
   * Drupal are converted to Entries and added.
   *
   * @param CredCom $credcom
   * @return Transaction
   * @throws CCFailure
   */
  static function convertRelayedTransactionToDrupal(CredCom $credcom, Currency $currency) : Transaction {
    $remote_wallet_id = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
    // Convert credcom entries into the transaction and child transactions.
    $props = static::convertCCEntryToDrupalValues($credcom->entries[0], $currency);
    $props['uuid'] = $credcom->uuid;
    $transaction = Transaction::create($props);
    foreach (array_slice($credcom->entries, 1) as $entry) {
      $props = static::convertCCEntryToDrupalValues($entry, $currency);
      try {
        $child = $transaction->createChild($props);
      }
      catch (\Exception $e) { }
    }
    foreach ($transaction->assemble() as $child) {
      if (static::isRemote($child)) {
        $e = static::ConvertTransactionToCcEntry($child);
        $credcom->entries[] = $e;
        $credcom->additional[] = $e;
      }
    }
    foreach ($transaction->validate() as $violation) {
      /** @var Symfony\Component\Validator\ConstraintViolation $violation */
      $message = 'Error in new transaction: '.$violation->getMessage();
      \Drupal::logger('mcapi_credcom')->debug('@message', ['@message' => $message]);
      throw new CCFailure($message);
    }
    return $transaction;
  }

  /**
   *
   * @param \stdClass $entry
   * @return array
   * @throws CCFailure
   */
  private static function convertCCEntryToDrupalValues($entry, Currency $currency) : array  {
    $wallet_id = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
    $credcom_settings = $currency->getThirdPartySettings('mcapi_credcom');
    $props = [
      'type' => 'remote',
      'payee' => self::convertCCaddrToWalletId($entry->payee, $credcom_settings['node_name']),
      'payer' => self::convertCCaddrToWalletId($entry->payer, $credcom_settings['node_name']),
      'curr_id' => $currency->id(),
      'quantity' => $entry->quant / $credcom_settings['rate'],
      'description' => $entry->description
    ];
    // One and only one wallet is the intertrading wallet.
    if ($props['payee'] == $wallet_id xor $props['payer'] == $wallet_id) {
      return $props;
    }
    throw new CCFailure('Cannot make remote transaction with '.$entry->payee.'('.$props['payee'].') and '.$entry->payer .'('.$props['payer'].')');
  }

  /**
   * Get the wallet ID from the credcom path.
   * @param string $cc_path
   *   A Credit Commons account path
   * @return int
   */
  private static function convertCCaddrToWalletId(string $cc_path, $node_name) : int {
    $parts = explode('/', $cc_path);
    $wallet_id = str_replace('-', ' ', end($parts));
    $entityTypeManager = \Drupal::entityTypeManager();
    if (!isset($wid) and is_numeric($wallet_id)) {
      // Assuming its not a float!
      $wid = intval($wallet_id);
    }
    elseif (self::credcomPathIsLocal($cc_path, $node_name)) {// according to the path this is local
      // Wallet name
      $wids = $entityTypeManager->getStorage('mcapi_wallet')
        ->getQuery()->condition('name', $wallet_id)
        ->accessCheck(FALSE)
        ->execute();
      if ($wids) {
        $wid = reset($wids);
      }
      // User email address
      elseif (filter_var($wallet_id, FILTER_VALIDATE_EMAIL)) {
        $users = $entityTypeManager->getStorage('user')->loadByProperties(['mail' => end($parts)]);
        $user = reset($users);
      }
      // User name
      else {
        $uids = $entityTypeManager->getStorage('user')
          ->getQuery()->condition('name', $wallet_id)
          ->accessCheck(FALSE)
          ->execute();
        if ($uids)  {
          $user = \Drupal\user\Entity\User::load(reset($uids));
        }
        if (isset($user)) {
          $wid = WalletStorage::firstWalletOfEntity($user);
        }
      }
    }
    if (isset($wid)) {
      return $wid;
    }
    return \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
  }

  /**
   * Entity builder callback.
   *
   * Replace the remote wallet with the intertrading wallet and build the credcom
   * pseudo property.
   */
  static function entityBuilder($entity_type_id, $entity, $form, $form_state) {
    if ($entity->type->target_id == 'remote') {
      $remote_wallet_id = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
      // the remote wallet still has a value of the account path.
      if (isset($entity->payer->target_id)) {// outgoing
        $entity->payee->target_id = $remote_wallet_id;
        static::$remoteAccPath = $entity->payee->value;
      }
      else  {
        $entity->payer->target_id = $remote_wallet_id;
        static::$remoteAccPath = $entity->payer->value;
      }
      // Drupal builds the entity in both the validation and submit phases of the form.
      // We save the remote result to save doing the request twice.
      if (!$form_state->get('credCom')) {
        // Remote build and validation.
        $curr_ids = (array)$form_state->getFormObject()
          ->getFormDisplay($form_state)
          ->getComponent('worth')['settings']['currencies'];
        $currency = Currency::load($curr_ids[0]);
        $credCom = new Credcom(
          xid: $entity->xid->value,
          uuid: $entity->uuid(),
          workflow: new Workflow($form_state->getValue('workflow')),
          entries: [static::ConvertTransactionValuesToCcEntry($form_state->getValues())],
          curr_id: $currency->id()
        );
        $requester = new TrunkwardNodeRequester($currency);
        try {
          [$new_rows, $valid_secs_left] = $requester->buildValidateTransaction($credCom->getTransaction());
        }
        catch (\CreditCommons\Exceptions\CCError $e) {
          $form_state->setErrorByName('', $e->getMessage() .' '. t('See error log for details'));
          \Drupal::logger('mcapi_credcom')->warning(
            'Failed to build transaction: @error',
            ['@error' => Markup::create('<pre>'.print_r($e, 1).'</pre>')]
          );
          return NULL;
        }
        foreach ($new_rows as $row) {
          // add the entry to credcom
          $credCom->entries[] = static::upcastEntry($row, $currency->getThirdPartySetting('mcapi_credcom', 'node_name'));
          // make a new transaction with the same serial number
          $props = Functions::convertCCEntryToDrupalValues($row, $currency);
          try {
            // add the new transaction to the cluster.
            $entity->createChild($props);
          }
          catch (\Exception $e) {}
        }
        $credCom->changeState(Workflow::STATE_VALIDATED);
        $form_state->set('credCom', $credCom);
      }
      $entity->credCom = $form_state->get('credCom');
    }
    return $entity;
  }

  /**
   * Get the last hash which should match the trunkward node's.
   */
  static function getLastHash($curr_id) : string {
    return \Drupal::database()
      ->select('mcapi_credcom_hashchain', 'cch')
      ->fields('cch', ['hash'])
      ->condition('curr_id', $curr_id)
      ->orderBy('id', 'desc')
      ->range(0, 1)
      ->execute()
      ->fetchField();
  }

  /**
   * @param Transaction $transaction
   * @return \CreditCommons\Entry
   */
  private static function ConvertTransactionToCcEntry(Transaction $transaction) : \CreditCommons\Entry {
    $credcom_settings = Currency::load($transaction->worth->curr_id)->getThirdpartySettings('mcapi_credcom');
    $data = (object)[
      'payer' => static::convertWalletToCredcomAcc($transaction->payer->entity, $credcom_settings['node_name']),
      'payee' => static::convertWalletToCredcomAcc($transaction->payee->entity, $credcom_settings['node_name']),
      'description' => $transaction->description->value,
      // Entry quants are always trunkward quants.
      'quant' => $transaction->worth->value * $credcom_settings['rate']
    ];
    return \CreditCommons\Entry::create($data);
  }

  /**
   * @param array $values
   * @return \CreditCommons\Entry
   */
  private static function ConvertTransactionValuesToCcEntry(array $values) : \CreditCommons\Entry {
    $curr_id = $values['worth'][0]['curr_id'];
    $settings = Currency::load($curr_id)->getThirdPartySettings('mcapi_credcom');
    $entry = [
      'description' => $values['description'][0]['value'],
      'quant' => $settings['rate'] * $values['worth'][0]['value'], // all quants are trunkward_quants
    ];
    if (isset($values['payer'][0]['value'])) { // payer is credcom wallet path
      $payee_wallet = Wallet::load($values['payee'][0]['target_id']);
      $entry['payee'] = $settings['node_name'] .'/'. str_replace(' ', '-', $payee_wallet->label());
      $entry['payer'] = $values['payer'][0]['value'];
    }
    else {
      $payer_wallet = Wallet::load($values['payer'][0]['target_id']);
      $entry['payer'] = $settings['node_name'] .'/'. str_replace(' ', '-', $payer_wallet->label());
      $entry['payee'] = $values['payee'][0]['value'];
    }
    $result =  static::upcastEntry((object)$entry, $settings['node_name']);
    return $result;
  }

  /**
   * @param Wallet $wallet
   * @return string
   * @todo this depends on upcastEntry or entitybuilder having run first.
   */
  private static function convertWalletToCredcomAcc(Wallet $wallet) : \CreditCommons\Account {
    if ($wallet->id() == \Drupal::config('mcapi_credcom.settings')->get('wallet_id')) {
      return new AccountRemote (id: static::$remoteAccPath);
    }
    else {
      return new Account(id: str_replace(' ', '-', $wallet->name->value, $node_name));
    }
  }

  /**
   * Utility, check if a credcom address refers to a local account, and if so
   * return the localised account name.
   *
   * @param string $cc_path
   * @param string $node_name
   * @return string
   *   the localised account path or empty if not recognised.
   */
  static function credcomPathIsLocal(string $cc_path, $node_name) : string {
    $parts = explode('/', $cc_path);
    $count = count($parts);
    if (count($parts) > 1 and $parts[$count-2] <> $node_name) {
      return '';
    }
    return str_replace('-', ' ', end($parts));
  }

  /**
   *
   * @param \stdClass $data
   * @param string $node_name
   * @return Entry
   */
  static function upcastEntry(\stdClass $data, $node_name) : Entry {
    foreach (['payee', 'payer'] as $role) {
      $path = $data->{$role};
      if ($local_name = static::credcomPathIsLocal($path, $node_name)) {
        $acc = new Account($local_name, $node_name);
      }
      else {
        if (!isset(static::$remoteAccPath)) {
          static::$remoteAccPath = $path;
        }
        $acc = new AccountRemote($path);
      }
      $data->{$role} = $acc;
    }
    return Entry::create($data);
  }

  /**
   * Exception handler to return credit commons errors to the network
   * Consider replacing with a class that extends Drupal\Core\Exception\ExceptionHandler
   * @param \Throwable $ex
   */
  static function CredcomExceptionHandler(\Throwable $ex) {
    $err = CCError::convertException($ex);
    \Drupal::logger('mcapi_credcom')
      ->error('@message', ['@message' => Markup::create('<pre>'.print_r($err, 1).'</pre>')]);
    header('Content-Type: application/json');
    header('HTTP/1.1: 500 Server error');
    print json_encode(['errors' => [$err]]);
    exit;
  }

  /**
   * Error handler to return credit commons errors to the network
   * @param \Throwable $ex
   */
  static function CredcomErrorHandler($severity, $message, $filename, $lineno) {
    $trace = array_slice(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS), 1);
    if ($severity > 8000) {// deprecation messages
      \Drupal::logger('mcapi_credcom')
      ->warning(
        'Deprecated: @message\n@break\n@trace',
        ['@message' => $message, '@break'=> "$filename:$lineno", '@trace' => Markup::Create('<pre>'.print_r($trace, 1).'</pre>')]
      );
    }
    else {
      $err = new CCFailure($message);
      //$err->trace = $trace;
      //$err->break = $filename . ':'. $lineno;
      \Drupal::logger('mcapi_credcom')
        ->error('@message', ['@message' =>  Markup::create('<pre>'.print_r($err, 1).'</pre>')]);
      header('Content-Type: application/json');
      header('HTTP/1.1: 500 Server error');
      print json_encode(['errors' => [$err]], 500);
      exit;
    }
  }


  static function currencyLookup($curr_id) : Currency {
    return Currency::load($curr_id);
    foreach (Currency::loadMultiple() as $curr) {
      if ($curr->getThirdPartySetting('mcapi_credcom', 'node_name') == $node_name) {
        return $curr;
      }
    }
    throw new \Exception("No currency corresponding to the node name $node_name in Functions::currencyLookup");
  }

  /**
   * Call this before using any code in the /CreditCommons namespace.
   * @global type $error_context
   * @param Currency $currency
   */
  static function createErrorContext(Currency $currency) {
    global $error_context;
    $method = $_SERVER['REQUEST_METHOD'];
    $path = $_SERVER['REQUEST_URI'];
    $trunkward_settings = $currency->getThirdPartySetting('mcapi_credcom', 'trunkward_node');
    $error_context = new \CreditCommons\ErrorContext(
      node: $currency->getThirdPartySetting('mcapi_credcom', 'node_name'),
      user: $trunkward_settings['name'],
      route_name: "drupalroute $method ".substr($path, 8) // Name of a credcom route preferably.
    );
  }

  /**
   * Use the given settings to create an account with the trunkward node.
   *
   * @param type $currency
   * @return bool
   */
  static function graft($currency) : bool {
    $trunkward_settings = $currency->getThirdPartySetting('mcapi_credcom', 'trunkward_node');
    $trunkward_url = $trunkward_settings['url']??'';
    Functions::createErrorContext($currency);
    $protocol = substr($_SERVER['SERVER_PROTOCOL'], 0, strpos($_SERVER['SERVER_PROTOCOL'], '/'));
    $current_site_url = strtolower($protocol) .'://'. $_SERVER['HTTP_HOST'] . '/credcom/'. $currency->id();
    $node_requester = new TrunkwardNodeRequester($currency);
    $abs_path = $node_requester->graft($current_site_url);
    if ($abs_path) {
      $currency->setThirdPartySetting('mcapi_credcom', 'absolute_path', $abs_path)->save();
    }
    return !empty($abs_path);
  }

}
