<?php

namespace Drupal\mcapi_credcom;
use Drupal\mcapi_credcom\Functions;
use Drupal\mcapi\Entity\Currency;
use Drupal\Core\Entity\Entity\EntityFormMode;
use CreditCommons\Exceptions\CCError;

/**
 * Class to handle outgoing requests to the trunkward node.
 * This is not a service because it must be initiated with different parameters sometimes.
 */
class TrunkwardNodeRequester extends \CreditCommons\NodeRequester {

  private string $nodeName;

  /**
   * Constructor
   */
  public function __construct(Currency $currency) {
    Functions::createErrorContext($currency);
    $remote_settings = $currency->getThirdPartySettings('mcapi_credcom');
    $this->nodeName = $remote_settings['node_name'];
    parent::__construct($remote_settings['trunkward_node']['url'], 4);
    $last_hash = Functions::getLastHash($currency->id());
    $this->authenticateAsNode(
      $this->nodeName,
      strlen($last_hash) ? $last_hash : $remote_settings['trunkward_node']['name']
    );
  }

  /**
   * {@inheritDoc}
   *
   * Theoretically this could be disintermediated by javascript calling the
   * trunkward node directly, but then the creds would have to be sent to client.
   */
  public function accountNameFilter(string $acc_path = '', $limit = 10) : array {
    $result = [];
    try {
      $result = parent::accountNameFilter($acc_path, $limit);
    }
    catch (\Throwable $ex) {
      \Drupal::logger('mcpai_credcom')->error(
        "accountNameFilter returned @type @message<br />@fullerror",
        ['@type' => get_class($ex), '@message' => $ex->getMessage(), '@fullerror'=> $ex]
      );
    }
    return $result;
  }

  public function convertPrice(string $node_path) : \stdClass {

  }

  private function checkFormMode($workflow) : void {
    $id = 'mcapi_transaction.remote_'.$workflow->id;
    if ($mode = EntityFormMode::load($id)) {
      return;
    }

    $label = t('Remote @name', ['@name' => $workflow->label]);
    $form_mode = EntityFormMode::create([
      'targetEntityType' => 'mcapi_transaction',
      'label' => $label,
      'id' => $id,
      'cache' => TRUE
    ]);
    if (isset($wf->creation->by['payer'])) {
      $firstparty = 'payer';
      $secondparty = 'payee';
      $verb = 'crediting';
      $template = t("{{ payer }}\r\nRemote account: {{ payee }}\r\nThey received {{ description }}\r\nI am billing them for {{ worth }}");
    }
    else {
      $firstparty = 'payee';
      $secondparty = 'payer';
      $verb = 'billing';
      $template = t("{{ payee }}\r\nRemote account: {{ payer }}\r\nI received {{ description }}\r\nI am crediting them for {{ worth }}");
    }

    $settings = [
      'title' => $label,
      'incoming' => TRUE,
      'path' => '/transact/remote/'.$workflow->id,
      'menu_title' => $workflow->label,
      'menu_weight'  => 5,
      'transaction_type'  => 'remote',
      'experience_twig'  => "{{ $firstparty }}\r\nRemote account: {{ $secondparty }}\r\nWhat was received {{ description }}\r\nI am $verb them for {{ worth }}",
      'experience_button'  => t('Send')
    ];
    foreach ($settings as $k => $v) {
      $form_mode->setThirdPartySetting('mcapi_forms', $k, $v);
    }
    $form_mode->save();
  }


  /**
   * @param string $url
   * @return bool
   */
  public function graft($endpoint_url) : string {
    try {
      $res = $this->setMethod('post')
        ->setBody(['acc_id' => $this->nodeName, 'url' => $endpoint_url])
        ->request("account");
    }
    catch(\CreditCommons\Exceptions\CCError $e) {
      if (str_contains($e->getMessage(), '403')) {
        \Drupal::messenger()->addError("The node at $this->baseUrl does not permit anonymous connections.");
      }
      elseif (str_contains($e->getMessage(), '404')) {
        \Drupal::messenger()->addError("This node is connected to $url.");
      }
    }
    return $res->abs_path??'';
  }


  protected function request(string $endpoint = '/') :\stdClass|NULL {
    try {
      $res =  parent::request($endpoint);
      return $res;
    }
    catch (CCError $e) {
      \Drupal::messenger()->addError($e->getMessage());
      throw $e;
    }
  }
}
