<?php

namespace Drupal\mcapi;

use Drupal\Core\Breadcrumb\Breadcrumb;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Breadcrumb\BreadcrumbBuilderInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Entity\Entity\EntityFormMode;

/**
 * Defines a class to build virtual breadcrumbs, i.e. not based on actual paths.
 *
 * @see \Drupal\Core\Breadcrumb\BreadcrumbBuilderInterface
 *
 * inject config
 */
class BreadcrumbBuilder implements BreadcrumbBuilderInterface {
  use StringTranslationTrait;

  /**
   * Routes and virtual paths. Admin paths not covered for now.
   */
  private $routes = [
    'view.mcapi_transactions.simple' =>            ['transactions'],
    'entity.mcapi_transaction.canonical' =>        ['transactions', 'id'],
    'entity.mcapi_transaction.add_form' =>         ['transactions', 'add'],
    'mcapi_mass.pay.12many' =>                     ['transactions', 'add', '12many'],
    'mcapi_mass.pay.many21' =>                     ['transactions', 'add', 'many21'],
    'mcapi.1stparty.bill' =>                       ['transactions', 'add', 'mcform'],
    'mcapi.1stparty.credit' =>                     ['transactions', 'add', 'mcform'],
    'view.mcapi_wallets.wallets_page' =>           ['wallets'],
    'entity.mcapi_wallet.canonical' =>             ['wallets', 'id'],
    'entity.mcapi_wallet.edit_form' =>             ['wallets', 'id', 'manage'],
    'entity.mcapi_wallet.delete_form' =>           ['wallets', 'id', 'delete'],
    'view.mcapi_wallet_transactions.wallet_page' =>['wallets', 'id', 'transactions'],
    'view.mcapi_income_expenditure.wallet_page' => ['wallets', 'id', 'inex'],
  //'entity.mcapi_transaction.add_remote' =>       ['transactions', 'add'],
  ];

  /**
   * {@inheritdoc}
   */
  public function applies(RouteMatchInterface $route_match) {
    $route_name = $route_match->getRouteName();
    return in_array($route_name, array_keys($this->routes)) or str_starts_with($route_name, 'entity.mcapi_transaction.');
  }

  /**
   * {@inheritdoc}
   */
  public function build(RouteMatchInterface $route_match) {
    $breadcrumb = new Breadcrumb();
    $links = [Link::createFromRoute($this->t('Home'), '<front>')];
    $route_name = $route_match->getRouteName();
    if (isset($this->routes[$route_name])) {
      $parts = $this->routes[$route_name];
    }
    elseif (str_starts_with($route_name, 'entity.mcapi_transaction.')) {
      $parts = ['transactions', 'add', 'transition'];
    }
    else {
      return;
    }

    switch($parts[0]) {
      case 'transactions':
        $links[] = new Link($this->t('Transactions'), Url::fromRoute('view.mcapi_transactions.simple'));
        break;
      case 'wallets':
        $links[] = new Link($this->t('Wallets'), Url::fromRoute('view.wallets.page_1'));
        break;
      case 'users':
        $user = $route_match->getParameter('user');
        $links[] = new Link($user->label(), Url::fromRoute('entity.user.canonical', ['user' => $user]));
        break;
    }

    if (isset($parts[1])) {
      switch ($parts[1]) {
        case 'add':
          if (empty($parts[2])) {
            $links[] = new Link($this->t("Create (as admin)"), Url::fromRoute('<current>', []));
            break;
          }
          switch ($parts[2]) {
            case '12many':
            case 'many21':
              $links[] = new Link($this->t('Mass transaction'), Url::fromRoute('<current>', []));
              break;
            case 'mcform':
              $route_parts = explode('.', $route_name);
              $mc_form = end($route_parts);
              $settings = EntityFormMode::load('mcapi_transaction.'.$mc_form)->getThirdpartySettings('mcapi_forms');
              $links[] = new Link($settings['title'], Url::fromRoute('<current>', []));
          }
          break;
        case 'id':
          if ($parts[0] == 'transactions') {
            $transaction = $route_match->getParameter('mcapi_transaction');
            if ($transaction->serial->value) {
              $links[] = new Link('#'.$transaction->serial->value, Url::fromRoute('entity.mcapi_transaction.canonical', ['mcapi_transaction' => $transaction->serial->value]));
            }
            else {
              $links[] = new Link($this->t('New'), Url::fromRoute('<current>'));
            }
          }
          elseif ($parts[0] == 'wallets') {
            $wallet = $route_match->getParameter('mcapi_wallet');
            $links[] = new Link($wallet->label(), Url::fromRoute('entity.mcapi_wallet.canonical', ['mcapi_wallet' => $wallet->id()]));
          }
          elseif ($parts[0] == 'users') {
            $user = $route_match->getParameter('user');
            $links[] = new Link($user->label(), Url::fromRoute('entity.user.canonical', ['mcapi_wallet' => $user->id()]));
          }
      }
      if (isset($parts[2])) {
        // these are all the last, so the link is always to the current page.
        switch($parts[2]) {
          case '12many':
            $links[] = new Link($this->t('One to many'), Url::fromRoute('<current>'));
            break;
          case 'many21':
            $links[] = new Link($this->t('Many to one'), Url::fromRoute('<current>'));
            break;
          case 'manage':
            $links[] = new Link($this->t('Edit'), Url::fromRoute('<current>'));
           break;
          case 'delete':
            $links[] = new Link($this->t('Delete'), Url::fromRoute('<current>'));
            break;
          case 'transactions':
            $links[] = new Link($this->t('Transactions'), Url::fromRoute('<current>'));
           break;
          case 'inex':
            $links[] = new Link($this->t('Statement'), Url::fromRoute('<current>'));
            break;
          case 'wallets':
            $links[] = new Link($this->t('Wallets'), Url::fromRoute('<current>'));
            break;
          case 'transitions':
            $links[] = new Link($this->t('Wallets'), Url::fromRoute('<current>'));
            break;
        }
      }
    }
    return $breadcrumb->setLinks($links)->addCacheContexts(['route']);
  }

}
